/**
 * Experiment - brute-force all MPU messages to find the corresponding properties.
 * 
 * Generated with:
 * for i in range(256):
 *    print("        { 0x06, 0x05, 0x01, 0x%02x, 0x%02x, 0x00 }," % (i, random.randint(1,255))) 
 * 
 * Usage:
 * 
 * - enable this in mpu.c, fix compiling errors
 * 
 * - slow down the communication to avoid overflows and keep the property processing serial if possible
 *   in mpu_send_next_spell, add some delay: eos_trigger_int(..., 100)
 * 
 * - enable property logging hooks in CAM/debugmsg.gdb (mpu_analyze_recv_data_log, prop_lookup_maybe_log)
 * 
 * - run the emulation like this (example for 6D):
     (
       ./run_canon_fw.sh 6D,firmware="boot=0" -s -S & 
        arm-none-eabi-gdb -x 6D/debugmsg.gdb -ex quit
     ) |& ansi2txt | grep "PropMgr:" | grep "MPU property" -A 5 | grep -E "MPU property|Analyze|lookup" \
        > mpu-bruteforce-6D.log
 *
 * - use the serial console to know when the command finished, or just wait for a couple of seconds
 *   (the GUI won't boot)
 * 
 * - if there are assertions or otherwise fatal errors, comment out the offending property and retry
 *   (usually it's the last one that was written in the log)
 * 
 * - command explained:
 *   - run under GDB to log property messages
 *   - strip ANSI color codes so they won't interfere with grep
 *   - the patterns we are looking for are:
 *     [     PropMgr:00039b6c ] MPU property: 03 2f 00 60 00 00 00 
 *     [     PropMgr:ff310138 ] prop_lookup_maybe 596cec 80030028 594510 4
 *     [     PropMgr:00039b6c ] MPU property: 01 f9 9d 00 
 *     [     PropMgr:ff3100c0 ] (01:16) NO AnalyzeMpuReceiveData 0x1 0xF9
 *   - there may be some noise (other messages) between the lines we are looking for
 *   - only look at messages from PropMgr task (ignore other tasks)
 *   - only look at "MPU property" messages and what follows them
 *   - from the messages that follow "MPU property", only keep the ones about Analyze and lookup
 *     (and, of course, the MPU property message)
 *   - when looking up properties, we will look at the first interesting line after "MPU property"
 *   - save the output to mpu-bruteforce-6D.log
 * 
 * - look up MPU properties in the brute force logs with:
 *      cat mpu-bruteforce-*.log | grep -A 1 "MPU property: 01 6f"
 * 
 *   result:
 *      [     PropMgr:0001796c ] MPU property: 01 6f 17 00 
 *      [     PropMgr:ff2e492c ] prop_lookup_maybe 80040034 80040034 513c10 1
 *   if you've brute-forced multiple camera models, you'll get multiple result - they should be consistent
 *   however, not all models have all properties - missing ones will look like this:
 *      [     PropMgr:ff1d85ac ] MPU property: 01 6f 17 00 
 *      [     PropMgr:ff1b7360 ] (01:16) NO AnalyzeMpuReceiveData 0x1 0x6F
 *
 * - look up property ID in a regular log (without brute-forcing) to identify it:
 *      ./run_canon_fw.sh 6D,firmware="boot=0" -d debugmsg |& ansi2txt | grep -i -C 5 80040034
 *   result:
 *      [     CtrlSrv:ff18f434 ] (84:01) copyDataToStorage eventID(0x80040034)Data(112)size(0)
 *      [     CtrlSrv:ff1946e8 ] (83:03) PROP_LIMITED_TV_VALUE_AT_AUTOISO (112)
 * 
 * - to confirm the MPU message 01 6f really maps to PROP_LIMITED_TV_VALUE_AT_AUTOISO, try changing its value
 *   in our case, in 6D.h - if missing, add it somewhere and make sure it gets sent
 *   for example: { 0x08, 0x07, 0x01, 0x6f, 0x13, 0x00, 0x00, 0x00 }
 *   result:
 *     [     CtrlSrv:ff18f434 ] (84:01) copyDataToStorage eventID(0x80040034)Data(19)size(0)
 *     [     CtrlSrv:ff1946e8 ] (83:03) PROP_LIMITED_TV_VALUE_AT_AUTOISO (19)
 * 
 * - to find/confirm a property name (doesn't always work):
 *   grep -nri "CtrlSrv.*copyDataToStorage.*80040034" startup-logs/ -A 1 | grep "CtrlSrv.*: PROP_" -B 1
 * 
 * -> hypothesis confirmed.
 * 
 * - Repeat for other property groups (0x02, 0x03, 0x04, 0x09 ... find and replace)
 * 
 * TODO: 
 * - remove the hardcoded spells
 * - run on all camera models
 */
static struct mpu_init_spell mpu_init_spells_bruteforce[] = {
    { { 0x06, 0x04, 0x02, 0x00, 0x00 }, .description = "Init", .out_spells = { /* spell #1 */
        { 0x06, 0x05, 0x01, 0x00, 0x19, 0x00 },
        { 0x06, 0x05, 0x01, 0x01, 0x36, 0x00 },
        { 0x06, 0x05, 0x01, 0x02, 0xb3, 0x00 },
        { 0x06, 0x05, 0x01, 0x03, 0xe1, 0x00 },
        { 0x06, 0x05, 0x01, 0x04, 0x23, 0x00 },
        { 0x06, 0x05, 0x01, 0x05, 0xd1, 0x00 },
        { 0x06, 0x05, 0x01, 0x06, 0xcc, 0x00 },
        { 0x06, 0x05, 0x01, 0x07, 0x8e, 0x00 },
        { 0x06, 0x05, 0x01, 0x08, 0x8b, 0x00 },
        { 0x06, 0x05, 0x01, 0x09, 0x48, 0x00 },
        { 0x06, 0x05, 0x01, 0x0a, 0xfc, 0x00 },
        { 0x06, 0x05, 0x01, 0x0b, 0x94, 0x00 },
        { 0x06, 0x05, 0x01, 0x0c, 0x71, 0x00 },
        { 0x06, 0x05, 0x01, 0x0d, 0xd9, 0x00 },
        { 0x06, 0x05, 0x01, 0x0e, 0x92, 0x00 },
        { 0x06, 0x05, 0x01, 0x0f, 0x43, 0x00 },
        { 0x06, 0x05, 0x01, 0x10, 0x2f, 0x00 },
        { 0x06, 0x05, 0x01, 0x11, 0xd9, 0x00 },
        { 0x06, 0x05, 0x01, 0x12, 0x2b, 0x00 },
        { 0x06, 0x05, 0x01, 0x13, 0x15, 0x00 },
        { 0x06, 0x05, 0x01, 0x14, 0xfc, 0x00 },
        { 0x06, 0x05, 0x01, 0x15, 0x03, 0x00 },
        { 0x06, 0x05, 0x01, 0x16, 0x8c, 0x00 },
        { 0x06, 0x05, 0x01, 0x17, 0x69, 0x00 },
        { 0x06, 0x05, 0x01, 0x18, 0x47, 0x00 },
        { 0x06, 0x05, 0x01, 0x19, 0xf6, 0x00 },
        { 0x06, 0x05, 0x01, 0x1a, 0xe5, 0x00 },
        { 0x06, 0x05, 0x01, 0x1b, 0x82, 0x00 },
        { 0x06, 0x05, 0x01, 0x1c, 0xbc, 0x00 },
        { 0x06, 0x05, 0x01, 0x1d, 0xb3, 0x00 },
        { 0x06, 0x05, 0x01, 0x1e, 0x08, 0x00 },
        { 0x06, 0x05, 0x01, 0x1f, 0xe2, 0x00 },
        { 0x06, 0x05, 0x01, 0x20, 0x8a, 0x00 },
        { 0x06, 0x05, 0x01, 0x21, 0xbf, 0x00 },
        { 0x06, 0x05, 0x01, 0x22, 0xb8, 0x00 },
        { 0x06, 0x05, 0x01, 0x23, 0x06, 0x00 },
        { 0x06, 0x05, 0x01, 0x24, 0x37, 0x00 },
        { 0x06, 0x05, 0x01, 0x25, 0x78, 0x00 },
        { 0x06, 0x05, 0x01, 0x26, 0x39, 0x00 },
        { 0x06, 0x05, 0x01, 0x27, 0x75, 0x00 },
        { 0x06, 0x05, 0x01, 0x28, 0xe7, 0x00 },
        { 0x06, 0x05, 0x01, 0x29, 0xfe, 0x00 },
        { 0x06, 0x05, 0x01, 0x2a, 0xd1, 0x00 },
        { 0x06, 0x05, 0x01, 0x2b, 0x74, 0x00 },
        { 0x06, 0x05, 0x01, 0x2c, 0x2f, 0x00 },
        { 0x06, 0x05, 0x01, 0x2d, 0x88, 0x00 },
        { 0x06, 0x05, 0x01, 0x2e, 0x74, 0x00 },
        { 0x06, 0x05, 0x01, 0x2f, 0x6b, 0x00 },
        { 0x06, 0x05, 0x01, 0x30, 0x5a, 0x00 },
        { 0x06, 0x05, 0x01, 0x31, 0x88, 0x00 },
        { 0x06, 0x05, 0x01, 0x32, 0xbb, 0x00 },
        { 0x06, 0x05, 0x01, 0x33, 0xc7, 0x00 },
        { 0x06, 0x05, 0x01, 0x34, 0x56, 0x00 },
        { 0x06, 0x05, 0x01, 0x35, 0x36, 0x00 },
        { 0x06, 0x05, 0x01, 0x36, 0x2a, 0x00 },
        { 0x06, 0x05, 0x01, 0x37, 0xf2, 0x00 },
        { 0x06, 0x05, 0x01, 0x38, 0x63, 0x00 },
        { 0x06, 0x05, 0x01, 0x39, 0x25, 0x00 },
        { 0x06, 0x05, 0x01, 0x3a, 0x90, 0x00 },
        { 0x06, 0x05, 0x01, 0x3b, 0x5f, 0x00 },
        { 0x06, 0x05, 0x01, 0x3c, 0x05, 0x00 },
        { 0x06, 0x05, 0x01, 0x3d, 0x46, 0x00 },
        { 0x06, 0x05, 0x01, 0x3e, 0xea, 0x00 },
        { 0x06, 0x05, 0x01, 0x3f, 0x28, 0x00 },
        { 0x06, 0x05, 0x01, 0x40, 0x3b, 0x00 },
        { 0x06, 0x05, 0x01, 0x41, 0x87, 0x00 },
        { 0x06, 0x05, 0x01, 0x42, 0x86, 0x00 },
        { 0x06, 0x05, 0x01, 0x43, 0xce, 0x00 },
        { 0x06, 0x05, 0x01, 0x44, 0xc2, 0x00 },
        { 0x06, 0x05, 0x01, 0x45, 0xf8, 0x00 },
        { 0x06, 0x05, 0x01, 0x46, 0x59, 0x00 },
        { 0x06, 0x05, 0x01, 0x47, 0x35, 0x00 },
        { 0x06, 0x05, 0x01, 0x48, 0xa8, 0x00 },
        { 0x06, 0x05, 0x01, 0x49, 0x7d, 0x00 },
        { 0x06, 0x05, 0x01, 0x4a, 0xd8, 0x00 },
        { 0x06, 0x05, 0x01, 0x4b, 0x99, 0x00 },
        { 0x06, 0x05, 0x01, 0x4c, 0x7d, 0x00 },
        { 0x06, 0x05, 0x01, 0x4d, 0x0f, 0x00 },
        { 0x06, 0x05, 0x01, 0x4e, 0x96, 0x00 },
        { 0x06, 0x05, 0x01, 0x4f, 0x1a, 0x00 },
        { 0x06, 0x05, 0x01, 0x50, 0x10, 0x00 },
        { 0x06, 0x05, 0x01, 0x51, 0x48, 0x00 },
        { 0x06, 0x05, 0x01, 0x52, 0xeb, 0x00 },
        { 0x06, 0x05, 0x01, 0x53, 0x58, 0x00 },
        { 0x06, 0x05, 0x01, 0x54, 0x81, 0x00 },
        { 0x06, 0x05, 0x01, 0x55, 0x41, 0x00 },
        { 0x06, 0x05, 0x01, 0x56, 0x5f, 0x00 },
        { 0x06, 0x05, 0x01, 0x57, 0x98, 0x00 },
        { 0x06, 0x05, 0x01, 0x58, 0x12, 0x00 },
        { 0x06, 0x05, 0x01, 0x59, 0xe7, 0x00 },
        { 0x06, 0x05, 0x01, 0x5a, 0xc8, 0x00 },
        { 0x06, 0x05, 0x01, 0x5b, 0x71, 0x00 },
        { 0x06, 0x05, 0x01, 0x5c, 0xdf, 0x00 },
        { 0x06, 0x05, 0x01, 0x5d, 0xc6, 0x00 },
        { 0x06, 0x05, 0x01, 0x5e, 0xb1, 0x00 },
        { 0x06, 0x05, 0x01, 0x5f, 0x41, 0x00 },
        { 0x06, 0x05, 0x01, 0x60, 0xff, 0x00 },
        { 0x06, 0x05, 0x01, 0x61, 0x9c, 0x00 },
        { 0x06, 0x05, 0x01, 0x62, 0x45, 0x00 },
        { 0x06, 0x05, 0x01, 0x63, 0x70, 0x00 },
        { 0x06, 0x05, 0x01, 0x64, 0xca, 0x00 },
        { 0x06, 0x05, 0x01, 0x65, 0x57, 0x00 },
        { 0x06, 0x05, 0x01, 0x66, 0x81, 0x00 },
        { 0x06, 0x05, 0x01, 0x67, 0x89, 0x00 },
        { 0x06, 0x05, 0x01, 0x68, 0xa5, 0x00 },
        { 0x06, 0x05, 0x01, 0x69, 0xc7, 0x00 },
        { 0x06, 0x05, 0x01, 0x6a, 0xa9, 0x00 },
        { 0x06, 0x05, 0x01, 0x6b, 0xc8, 0x00 },
        { 0x06, 0x05, 0x01, 0x6c, 0x25, 0x00 },
        { 0x06, 0x05, 0x01, 0x6d, 0x28, 0x00 },
        { 0x06, 0x05, 0x01, 0x6e, 0x12, 0x00 },
        { 0x06, 0x05, 0x01, 0x6f, 0x0a, 0x00 },
        { 0x06, 0x05, 0x01, 0x70, 0x77, 0x00 },
        { 0x06, 0x05, 0x01, 0x71, 0x64, 0x00 },
        { 0x06, 0x05, 0x01, 0x72, 0x69, 0x00 },
        { 0x06, 0x05, 0x01, 0x73, 0x93, 0x00 },
        { 0x06, 0x05, 0x01, 0x74, 0x48, 0x00 },
        { 0x06, 0x05, 0x01, 0x75, 0x34, 0x00 },
        { 0x06, 0x05, 0x01, 0x76, 0x84, 0x00 },
        { 0x06, 0x05, 0x01, 0x77, 0xa8, 0x00 },
        { 0x06, 0x05, 0x01, 0x78, 0xaf, 0x00 },
        { 0x06, 0x05, 0x01, 0x79, 0x24, 0x00 },
        { 0x06, 0x05, 0x01, 0x7a, 0xe2, 0x00 },
        { 0x06, 0x05, 0x01, 0x7b, 0xe6, 0x00 },
        { 0x06, 0x05, 0x01, 0x7c, 0x4b, 0x00 },
        { 0x06, 0x05, 0x01, 0x7d, 0x1b, 0x00 },
        { 0x06, 0x05, 0x01, 0x7e, 0x46, 0x00 },
        { 0x06, 0x05, 0x01, 0x7f, 0x3e, 0x00 },
        { 0x06, 0x05, 0x01, 0x80, 0xef, 0x00 },
        { 0x06, 0x05, 0x01, 0x81, 0x3c, 0x00 },
        { 0x06, 0x05, 0x01, 0x82, 0xe2, 0x00 },
        { 0x06, 0x05, 0x01, 0x83, 0x47, 0x00 },
        { 0x06, 0x05, 0x01, 0x84, 0xbe, 0x00 },
        { 0x06, 0x05, 0x01, 0x85, 0x6f, 0x00 },
        { 0x06, 0x05, 0x01, 0x86, 0x8b, 0x00 },
        { 0x06, 0x05, 0x01, 0x87, 0x3c, 0x00 },
        { 0x06, 0x05, 0x01, 0x88, 0x51, 0x00 },
        { 0x06, 0x05, 0x01, 0x89, 0x08, 0x00 },
        { 0x06, 0x05, 0x01, 0x8a, 0xf9, 0x00 },
        { 0x06, 0x05, 0x01, 0x8b, 0xa8, 0x00 },
        { 0x06, 0x05, 0x01, 0x8c, 0xb9, 0x00 },
        { 0x06, 0x05, 0x01, 0x8d, 0x11, 0x00 },
        { 0x06, 0x05, 0x01, 0x8e, 0x6c, 0x00 },
        { 0x06, 0x05, 0x01, 0x8f, 0x1a, 0x00 },
        { 0x06, 0x05, 0x01, 0x90, 0xee, 0x00 },
        { 0x06, 0x05, 0x01, 0x91, 0x5e, 0x00 },
        { 0x06, 0x05, 0x01, 0x92, 0x45, 0x00 },
        { 0x06, 0x05, 0x01, 0x93, 0xd7, 0x00 },
        { 0x06, 0x05, 0x01, 0x94, 0x1d, 0x00 },
        { 0x06, 0x05, 0x01, 0x95, 0x5f, 0x00 },
        { 0x06, 0x05, 0x01, 0x96, 0x95, 0x00 },
        { 0x06, 0x05, 0x01, 0x97, 0x48, 0x00 },
        { 0x06, 0x05, 0x01, 0x98, 0xdb, 0x00 },
        { 0x06, 0x05, 0x01, 0x99, 0x74, 0x00 },
        { 0x06, 0x05, 0x01, 0x9a, 0x84, 0x00 },
        { 0x06, 0x05, 0x01, 0x9b, 0xf9, 0x00 },
        { 0x06, 0x05, 0x01, 0x9c, 0x2d, 0x00 },
        { 0x06, 0x05, 0x01, 0x9d, 0x0e, 0x00 },
        { 0x06, 0x05, 0x01, 0x9e, 0x5d, 0x00 },
        { 0x06, 0x05, 0x01, 0x9f, 0xd7, 0x00 },
        { 0x06, 0x05, 0x01, 0xa0, 0xa8, 0x00 },
        { 0x06, 0x05, 0x01, 0xa1, 0xc5, 0x00 },
        { 0x06, 0x05, 0x01, 0xa2, 0x94, 0x00 },
        { 0x06, 0x05, 0x01, 0xa3, 0xac, 0x00 },
        { 0x06, 0x05, 0x01, 0xa4, 0x98, 0x00 },
        { 0x06, 0x05, 0x01, 0xa5, 0xd7, 0x00 },
        { 0x06, 0x05, 0x01, 0xa6, 0x78, 0x00 },
        { 0x06, 0x05, 0x01, 0xa7, 0x87, 0x00 },
        { 0x06, 0x05, 0x01, 0xa8, 0x42, 0x00 },
        { 0x06, 0x05, 0x01, 0xa9, 0x39, 0x00 },
        { 0x06, 0x05, 0x01, 0xaa, 0xe1, 0x00 },
        { 0x06, 0x05, 0x01, 0xab, 0x42, 0x00 },
        { 0x06, 0x05, 0x01, 0xac, 0xa3, 0x00 },
        { 0x06, 0x05, 0x01, 0xad, 0x0a, 0x00 },
        { 0x06, 0x05, 0x01, 0xae, 0x0a, 0x00 },
        { 0x06, 0x05, 0x01, 0xaf, 0xbf, 0x00 },
        { 0x06, 0x05, 0x01, 0xb0, 0x42, 0x00 },
        { 0x06, 0x05, 0x01, 0xb1, 0xde, 0x00 },
        { 0x06, 0x05, 0x01, 0xb2, 0x76, 0x00 },
        { 0x06, 0x05, 0x01, 0xb3, 0xbf, 0x00 },
        { 0x06, 0x05, 0x01, 0xb4, 0xe2, 0x00 },
        { 0x06, 0x05, 0x01, 0xb5, 0xac, 0x00 },
        { 0x06, 0x05, 0x01, 0xb6, 0x9a, 0x00 },
        { 0x06, 0x05, 0x01, 0xb7, 0x13, 0x00 },
        { 0x06, 0x05, 0x01, 0xb8, 0x62, 0x00 },
        { 0x06, 0x05, 0x01, 0xb9, 0x5a, 0x00 },
        { 0x06, 0x05, 0x01, 0xba, 0x70, 0x00 },
        { 0x06, 0x05, 0x01, 0xbb, 0x15, 0x00 },
        { 0x06, 0x05, 0x01, 0xbc, 0xf9, 0x00 },
        { 0x06, 0x05, 0x01, 0xbd, 0x6b, 0x00 },
        { 0x06, 0x05, 0x01, 0xbe, 0x2c, 0x00 },
        { 0x06, 0x05, 0x01, 0xbf, 0x71, 0x00 },
        { 0x06, 0x05, 0x01, 0xc0, 0xda, 0x00 },
        { 0x06, 0x05, 0x01, 0xc1, 0x3a, 0x00 },
        { 0x06, 0x05, 0x01, 0xc2, 0x26, 0x00 },
        { 0x06, 0x05, 0x01, 0xc3, 0xd5, 0x00 },
        { 0x06, 0x05, 0x01, 0xc4, 0xa6, 0x00 },
        { 0x06, 0x05, 0x01, 0xc5, 0x3e, 0x00 },
        { 0x06, 0x05, 0x01, 0xc6, 0xcb, 0x00 },
        { 0x06, 0x05, 0x01, 0xc7, 0xb6, 0x00 },
        { 0x06, 0x05, 0x01, 0xc8, 0xec, 0x00 },
        { 0x06, 0x05, 0x01, 0xc9, 0x06, 0x00 },
        { 0x06, 0x05, 0x01, 0xca, 0x0c, 0x00 },
        { 0x06, 0x05, 0x01, 0xcb, 0xcc, 0x00 },
        { 0x06, 0x05, 0x01, 0xcc, 0xf2, 0x00 },
        { 0x06, 0x05, 0x01, 0xcd, 0xa3, 0x00 },
        { 0x06, 0x05, 0x01, 0xce, 0x60, 0x00 },
        { 0x06, 0x05, 0x01, 0xcf, 0x21, 0x00 },
        { 0x06, 0x05, 0x01, 0xd0, 0xea, 0x00 },
        { 0x06, 0x05, 0x01, 0xd1, 0x81, 0x00 },
        { 0x06, 0x05, 0x01, 0xd2, 0x4b, 0x00 },
        { 0x06, 0x05, 0x01, 0xd3, 0x59, 0x00 },
        { 0x06, 0x05, 0x01, 0xd4, 0x89, 0x00 },
        { 0x06, 0x05, 0x01, 0xd5, 0xc4, 0x00 },
        { 0x06, 0x05, 0x01, 0xd6, 0x53, 0x00 },
        { 0x06, 0x05, 0x01, 0xd7, 0x72, 0x00 },
        { 0x06, 0x05, 0x01, 0xd8, 0x51, 0x00 },
        { 0x06, 0x05, 0x01, 0xd9, 0x58, 0x00 },
        { 0x06, 0x05, 0x01, 0xda, 0x52, 0x00 },
        { 0x06, 0x05, 0x01, 0xdb, 0x98, 0x00 },
        { 0x06, 0x05, 0x01, 0xdc, 0xa1, 0x00 },
        { 0x06, 0x05, 0x01, 0xdd, 0x1c, 0x00 },
        { 0x06, 0x05, 0x01, 0xde, 0xf7, 0x00 },
        { 0x06, 0x05, 0x01, 0xdf, 0xa3, 0x00 },
        { 0x06, 0x05, 0x01, 0xe0, 0x9d, 0x00 },
        { 0x06, 0x05, 0x01, 0xe1, 0x41, 0x00 },
        { 0x06, 0x05, 0x01, 0xe2, 0xa6, 0x00 },
        { 0x06, 0x05, 0x01, 0xe3, 0x2d, 0x00 },
        { 0x06, 0x05, 0x01, 0xe4, 0x24, 0x00 },
        { 0x06, 0x05, 0x01, 0xe5, 0x94, 0x00 },
        { 0x06, 0x05, 0x01, 0xe6, 0x51, 0x00 },
        { 0x06, 0x05, 0x01, 0xe7, 0xb3, 0x00 },
        { 0x06, 0x05, 0x01, 0xe8, 0x34, 0x00 },
        { 0x06, 0x05, 0x01, 0xe9, 0xcb, 0x00 },
        { 0x06, 0x05, 0x01, 0xea, 0x4b, 0x00 },
        { 0x06, 0x05, 0x01, 0xeb, 0xbb, 0x00 },
        { 0x06, 0x05, 0x01, 0xec, 0xb9, 0x00 },
        { 0x06, 0x05, 0x01, 0xed, 0xd2, 0x00 },
        { 0x06, 0x05, 0x01, 0xee, 0x98, 0x00 },
        { 0x06, 0x05, 0x01, 0xef, 0x2c, 0x00 },
        { 0x06, 0x05, 0x01, 0xf0, 0x72, 0x00 },
        { 0x06, 0x05, 0x01, 0xf1, 0xf9, 0x00 },
        { 0x06, 0x05, 0x01, 0xf2, 0xe4, 0x00 },
        { 0x06, 0x05, 0x01, 0xf3, 0xf3, 0x00 },
        { 0x06, 0x05, 0x01, 0xf4, 0xa1, 0x00 },
        { 0x06, 0x05, 0x01, 0xf5, 0x77, 0x00 },
        { 0x06, 0x05, 0x01, 0xf6, 0x45, 0x00 },
        { 0x06, 0x05, 0x01, 0xf7, 0xbe, 0x00 },
        { 0x06, 0x05, 0x01, 0xf8, 0xf6, 0x00 },
        { 0x06, 0x05, 0x01, 0xf9, 0x9d, 0x00 },
        { 0x06, 0x05, 0x01, 0xfa, 0xc3, 0x00 },
        { 0x06, 0x05, 0x01, 0xfb, 0x57, 0x00 },
        { 0x06, 0x05, 0x01, 0xfc, 0x79, 0x00 },
        { 0x06, 0x05, 0x01, 0xfd, 0x50, 0x00 },
        { 0x06, 0x05, 0x01, 0xfe, 0x3e, 0x00 },
      //{ 0x06, 0x05, 0x01, 0xff, 0xd8, 0x00 },
        { 0 } } },
};
